"""
Poll code
"""

import logging
from functools import partial

import anyio
from asyncscope import scope
from moat.util import attrdict, merge, to_attrdict

from ..client import ModbusClient
from .device import ServerDevice, ClientDevice, fixup
from ..server import ModbusServer, SerialModbusServer

logger = logging.getLogger(__name__)

async def dev_poll(cfg, mt_kv, *, task_status=None):
    """
    Run a device task on this set of devices, as configured by the config.

    The config will be preprocessed by `moat.modbus.dev.device.fixup`; the
    result will be returned via @task_status after setup is complete.
    """
    sl = cfg.setdefault("slots", attrdict())
    sl.setdefault("write", attrdict())
    sl._apply_default = True  # pylint:disable=protected-access
    cfg = fixup(cfg)

    s = cfg.setdefault("src", attrdict())
    sl = cfg.slots

    async with ModbusClient() as cl, anyio.create_task_group() as tg:
        nd = 0

        async def make_dev(v, Reg, **kw):
            kw = to_attrdict(kw)
            vs = v.setdefault("src", attrdict())
            merge(vs, kw, replace=False)
            vsl = v.setdefault("slots", attrdict())
            merge(vsl, sl, replace=False)

            logger.info("Starting %r", vs)

            dev = ClientDevice(client=cl, factory=Reg)
            await dev.load(data=v)

            return await scope.spawn_service(dev.as_scope)

        if mt_kv is None:
            from .device import Register as Reg  # pylint: disable=import-outside-toplevel
            RegS = Reg
        else:
            # The MoaT-KV client must live longer than the taskgroup
            from .kv import Register  # pylint: disable=import-outside-toplevel

            Reg = partial(Register, mt_kv=mt_kv, tg=tg)  # noqa: F811
            RegS = partial(Register, mt_kv=mt_kv, tg=tg, is_server=True)  # noqa: F811

        # relay-out server(s)
        servers = []
        for s in cfg.get("server", ()):
            if "serial" in s:
                port = s.get("port", None)
                kw = s["serial"]
                if port is not None:
                    kw["port"] = port
                srv = SerialModbusServer(**kw)
            elif "host" in s or ("port" in s and isinstance(s["port"],int)):
                kw = {}
                for k,v in s.items():
                    if isinstance(k,str):
                        if k == "units":
                            continue
                        if k == "host":
                            k = "address"
                        kw[k] = v
                srv = ModbusServer(**kw)
            else:
                raise ValueError("neither serial nor TCP config found")
            servers.append(srv)

            for u,v in s.get("units",{}).items():
                dev = ServerDevice(factory=RegS)
                await dev.load(data=v)
                srv.add_unit(u, dev)
                nd += 1


        def do_attach(v, dev):
            p = v.get("server", None)
            if p is None:
                return
            if isinstance(p,int):
                s,u = servers[p//1000],p%1000
            else:
                s,u = servers[p[0]],p[1]
            s.add_unit(u, dev.unit)

            nonlocal nd
            nd += 1

        # serial clients
        for h, hv in cfg.get("ports", {}).items():
            try:
                sp = hv["serial"]
            except KeyError:
                logger.error("No serial params for port %r", h)
                continue
            for u, v in hv.items():
                if not isinstance(u, int):
                    continue
                dev = await make_dev(v, Reg, port=h, serial=sp, unit=u)
                tg.start_soon(dev.poll)
                do_attach(v,dev)

        # TCP clients
        for h, hv in cfg.get("hosts", {}).items():
            for u, v in hv.items():
                if not isinstance(u, int):
                    continue
                dev = await make_dev(v, Reg, host=h, unit=u)
                tg.start_soon(dev.poll)
                do_attach(v,dev)

        # more TCP clients
        for h, hv in cfg.get("hostports", {}).items():
            for p, pv in hv.items():
                if not isinstance(p, int):
                    continue
                for u, v in pv.items():
                    dev = await make_dev(v, Reg, host=h, port=p, unit=u)
                    tg.start_soon(dev.poll)
                    do_attach(v,dev)

        for s in servers:
            evt = anyio.Event()
            tg.start_soon(partial(s.serve, opened=evt))
            await evt.wait()

        if task_status is not None:
            task_status.started(cfg)

        if nd:
            logger.info("Running.")
        if not nd:
            logger.error("No devices to poll found.")

        pass  # wait until all tasks are done



"""
Basic client/server tests
"""
import pytest

from moat.modbus.client import ModbusClient
from moat.modbus.server import ModbusServer
from moat.modbus.types import FloatValue, HoldingRegisters, LongValue, StringValue


@pytest.mark.anyio
async def test_rw():
    """Your basic client/server read/write test"""
    async with ModbusServer(address="127.0.0.1", port=0) as srv:
        sru = srv.add_unit(12)
        srv1 = sru.add(HoldingRegisters, 34, LongValue(123456))
        srv2 = sru.add(HoldingRegisters, 36, FloatValue(9999.125))
        srv3 = sru.add(HoldingRegisters, 38, StringValue(10, "hélþ"))

        async with ModbusClient() as cli:
            async with cli.host("127.0.0.1", srv.port) as clh, clh.unit(12) as clu, clu.slot(
                "x"
            ) as cls:
                clv1 = cls.add(HoldingRegisters, 34, LongValue())
                clv2 = cls.add(HoldingRegisters, 36, FloatValue())
                clv3 = cls.add(HoldingRegisters, 38, StringValue(10))

                await cls._getValues()  # pylint:disable=protected-access
                assert clv1.value == 123456
                assert clv2.value == 9999.125
                assert clv3.value == "hélþ"

                c = sru.store["h"].changed
                assert not c.is_set()
                clv1._value_w = 345678
                clv2._value_w = clv2._value * 3
                clv3._value_w = "CoMiNg"
                await cls._setValues()  # pylint:disable=protected-access
                assert srv1.value == 345678
                assert srv2.value == 9999.125 * 3
                assert srv3.value == "CoMiNg"
                assert c.is_set()



import logging
from time import time

import asyncclick as click
import pytest
import trio
import os
from moat.src.test import raises
from moat.util import P, PathLongener, yload

from moat.kv.client import ServerError
from moat.kv.mock.mqtt import stdtest
from moat.modbus.dev.poll import dev_poll
from moat.modbus.types import HoldingRegisters, IntValue
from moat.modbus.client import ModbusClient

_lg = logging.getLogger("moat.mqtt.mqtt.protocol.handler")
_lg.level = logging.WARNING

_lg = logging.getLogger("moat.kv.server.test_0")
_lg.level = logging.WARNING

logger = logging.getLogger(__name__)

cfg1_ = """
server:
- host: 127.0.0.1
  port: -1
  units:
    32:
      regs:
        no_question:
          reg_type: h
          register: 12342
          dest: !P a.srv.dst
          src: !P a.srv.src
          type: uint
          len: 1
"""

cfg2_ = """
slots:
  1sec:
    read_delay: 1
    read_align: false
hostports:
  localhost:
    PORT:
      32:
        regs:
          the_answer:
            reg_type: h
            register: 12342
            type: uint
            len: 1
            dest: !P a.cli.dst
            src: !P a.cli.src
            slot: 1sec
    
  
"""

@pytest.mark.trio
async def test_kv_poll(autojump_clock):  # pylint: disable=unused-argument
    cfg1 = yload(cfg1_, attr=True)
    cfg2 = yload(cfg2_, attr=True)
    PORT = 40000 + (os.getpid() + 20) % 10000
    cfg1.server[0].port = PORT
    cfg2.hostports.localhost[PORT] = cfg2.hostports.localhost.PORT
    del cfg2.hostports.localhost.PORT

    async with (
            stdtest(args={"init": 123}, tocks=50) as st,
            st.client() as c,
            trio.open_nursery() as tg,
        ):
        assert (await c.get(P(":"))).value == 123
        await c.set(P("a.srv.src"), value=42)
        cfg1 = await tg.start(dev_poll, cfg1, c)
        reg = cfg1.server[0].units[32].regs.no_question
        await trio.sleep(1)
        assert reg.value_w == 42

        await c.set(P("a.srv.src"), value=44)
        await trio.sleep(1)
        assert reg.value_w == 44

        reg.set(43)
        await trio.sleep(2)

        rv = await c.get(P("a.srv.dst"))
        #assert rv == 43
        #assert reg.value_w == 44
        #assert reg.value == 43

        tg.cancel_scope.cancel()
        return # owch

        async with (
            ModbusClient() as g,
            g.host("localhost", PORT) as h,
            h.unit(32) as u,
            u.slot("default") as s,
        ):
            v = s.add(HoldingRegisters, 12342, IntValue)
            res = await s._getValues()
            pass # v,res


        pass # ex






"""
Device file handling, with a basic multi-device client
"""

import logging
from functools import partial
from pathlib import Path as FSPath

import anyio
import asyncclick as click
from moat.util import attrdict, yload, yprint

from ..client import ModbusClient
from .device import ClientDevice, fixup
from .poll import dev_poll

logger = logging.getLogger(__name__)


@click.group()
def cli():
    """Modbus device polling"""
    pass


@cli.command()
@click.option("-r", "--raw", is_flag=True, help="don't postprocess")
@click.option("-R", "--no-refs", is_flag=True, help="don't process references")
@click.argument("path", type=click.Path("r"))
def dump(path, raw, no_refs):
    """Dump a postprocessed file"""
    path = FSPath(path)
    with path.open("r") as f:
        d = yload(f, attr=True)
    if not raw:
        d = fixup(d, do_refs=not no_refs, this_file=path)
    yprint(d)


@cli.command()
@click.argument("path", type=click.File("r"))
@click.pass_context
async def poll(ctx, path):
    """Poll Modbus devices as directed via YAML."""
    obj = ctx.obj

    cfg = yload(path, attr=True)

    if "kv" in obj.cfg:
        # pylint: disable=import-outside-toplevel
        from moat.kv.client import client_scope

        mt_kv = await client_scope(**obj.cfg.kv)
    else:
        mt_kv = None

    await dev_poll(cfg, mt_kv)



#!/usr/bin/python3

"""
Modbus server classes for serial(RTU) and TCP.
"""

import logging
import socket
from binascii import b2a_hex
from contextlib import asynccontextmanager
from typing import Type, Union
import time

import anyio
from anyio.abc import SocketAttribute
from moat.util import CtxObj
from pymodbus.constants import Defaults
from pymodbus.datastore import ModbusServerContext, ModbusSlaveContext
from pymodbus.device import ModbusControlBlock, ModbusDeviceIdentification
from pymodbus.exceptions import NoSuchSlaveException
from pymodbus.factory import ServerDecoder
from pymodbus.pdu import ModbusExceptions as merror, ExceptionResponse
from pymodbus.utilities import hexlify_packets

from moat.modbus.types import BaseValue, DataBlock, TypeCodec

_logger = logging.getLogger(__name__)


class UnitContext(ModbusSlaveContext):
    """
    This module implements a slave context for servers that stores
    individual variables.
    """

    def __init__(self, server=None, unit=None):
        super().__init__(
            di=DataBlock(),
            co=DataBlock(),
            ir=DataBlock(),
            hr=DataBlock(),
            zero_mode=True,
        )
        if server:
            self.unit = unit
            server._add_unit(self)

    def add(
        self, typ: TypeCodec, offset: int, val: Union[BaseValue, Type[BaseValue]]
    ) -> BaseValue:
        """Add a field to be served.

        :param typ: The `TypeCodec` instance to use.
        :param offset: The value's numeric offset, zero-based.
        :param val: The data type (baseValue instance)

        `cls` is either the decoder (subclass of `BaseValue`),
        or an existing `BaseValue` instance.
        """
        k = self.store[typ.key]
        if isinstance(val, type):
            val = val()
        k.add(offset, val)
        return val

    def remove(self, typ: TypeCodec, offset: int):
        """Remove a field to be requested.

        :param typ: the `TypeCodec` to use
        :param offset: the offset where the value is located

        Returns the field in question, or none if it doesn't exist.
        """
        k = self.store[typ.key]
        return k.delete(offset)


class BaseModbusServer(CtxObj):
    """Basic base class for servers."""

    def __init__(self, identity=None, response_manipulator=None):
        self.context = ModbusServerContext(single=False)
        self.context._slaves = self.units = {}
        self.control = ModbusControlBlock()
        self.broadcast_enable = False
        self.response_manipulator = response_manipulator

        if isinstance(identity, ModbusDeviceIdentification):
            self.control.Identity.update(identity)

        if identity is None:
            identity = ModbusDeviceIdentification()
            identity.VendorName = "Matthias Urlichs"
            identity.ProductCode = "MoaT.modbus"
            identity.VendorUrl = "http://M-o-a-T.org/"
            identity.ProductName = "MoaT-Modbus Test"
            identity.ModelName = "MoaT-Modbus Test"
            identity.MajorMinorRevision = "1.0"
        self.identity = identity

    def add_unit(self, unit, ctx=None) -> UnitContext:
        """
        Add an empty unit (= slave context) to this server (and return it).

        The unit must not exist.
        """
        if unit in self.units:
            raise RuntimeError(f"Unit {unit} already exists")
        if ctx is None:
            return UnitContext(self, unit)
        self.units[unit] = ctx

    def _add_unit(self, unit):
        self.units[unit.unit] = unit

    async def serve(self, opened=None):
        """The actual server. Override me."""
        raise RuntimeError("You need to override .serve")

    async def process_request(self, request):
        """Basic request processor"""
        context = self.context[request.unit_id]
        if hasattr(context, "process_request"):
            response = await context.process_request(request)
        else:
            response = request.execute(context)
        return response

    @asynccontextmanager
    async def _ctx(self):
        async with anyio.create_task_group() as tg:
            evt = anyio.Event()
            tg.start_soon(self.serve, evt)
            try:
                await evt.wait()
                yield self
            finally:
                tg.cancel_scope.cancel()


class SerialModbusServer(BaseModbusServer):
    """
    A simple serial Modbus server (RTU).
    """

    _serial = None
    framer = None
    ignore_missing_slaves = False
    single = False

    def __init__(self, identity=None, timeout=None, **args):
        super().__init__(identity=identity)
        self.args = args
        self.timeout = timeout

        from pymodbus.framer.rtu_framer import (  # pylint: disable=import-outside-toplevel
            ModbusRtuFramer,
        )
        class Framer(ModbusRtuFramer):
            def _validate_unit_id(self, unit, single):
                return True

        self.decoder = ServerDecoder()  # pylint: disable=no-value-for-parameter ## duh?
        self.Framer = Framer

    async def serve(self, opened=None):
        from anyio_serial import Serial  # pylint: disable=import-outside-toplevel

        async with Serial(**self.args) as ser:
            self._serial = ser
            self.framer = self.Framer(self.decoder)

            if opened is not None:
                opened.set()
            t = time.monotonic()
            while True:
                with anyio.move_on_after(0.1):
                    await ser.receive()
                    break
            while True:
                if self.timeout:
                    with anyio.fail_after(self.timeout):
                        data = await ser.receive()
                else:
                    data = await ser.receive()
                t2 = time.monotonic()
                if t2-t > 0.2:
                    self.framer.resetFrame()
                t = t2
                msgs = []
                self.framer.processIncomingPacket(data=data, unit=0, callback=msgs.append, single=True)
                for msg in msgs:
                    with anyio.fail_after(2):
                        await self._process(msg)

    async def _process(self, request):
        broadcast = False
        unit = request.unit_id
        tid = request.transaction_id

        try:
            if self.broadcast_enable and not request.unit_id:
                broadcast = True
                # if broadcasting then execute on all slave contexts,
                # note response will be ignored
                for unit_id in self.context.slaves():
                    response = await request.execute(self.context[unit_id])
            else:
                response = await self.process_request(request)
        except NoSuchSlaveException:
            txt = f"requested slave does not exist: {request.unit_id}"
            _logger.error(txt)
            if self.ignore_missing_slaves:
                return  # the client will simply timeout waiting for a response
            response = request.doException(merror.GatewayNoResponse)
        except Exception:  # pylint: disable=broad-except
            _logger.exception("Unable to fulfill request")
            response = request.doException(merror.SlaveFailure)
        # no response when broadcasting
        response.unit_id = unit
        response.transaction_id = tid
        if isinstance(response,ExceptionResponse):
            _logger.error("Source: %r %d %d %d", type(request), request.function_code, request.address, getattr(request,"count",1))

        if response.should_respond and not broadcast:
            response.transaction_id = request.transaction_id
            response.unit_id = request.unit_id
            skip_encoding = False
            if self.response_manipulator:
                response, skip_encoding = self.response_manipulator(response)
            if not skip_encoding:
                response = self.framer.buildPacket(response)
#           if _logger.isEnabledFor(logging.DEBUG):
#               _logger.debug("send: [%s]- %s", request, b2a_hex(response))

            await self._serial.send(response)


class RelayServer:
    """
    A mix-in class to teach a server to forward all requests to a client
    """

    single = True

    def __init__(self, client, *a, **k):
        self._client = client
        super().__init__(*a, **k)

    async def _process(self, request):
        request = self.mon_request(request)
        tid = request.transaction_id
        resp = await self._client.execute(request)

        resp.transaction_id = tid
        resp = self.mon_response(resp) or resp
        resp = self.framer.buildPacket(resp)  # pylint:disable=no-member
        await self._serial.send(resp)  # pylint:disable=no-member

    def mon_request(self, request):
        """Request monitor. Override me."""
        return request

    def mon_response(self, response):
        """Response monitor. Override me."""
        return response


class ModbusServer(BaseModbusServer):
    """TCP Modbus server.

    If the identity structure is not passed in, the ModbusControlBlock
    uses its own empty structure.

    :param identity: An optional identity structure
    :param address: An optional address to bind to.
    :param port: the TCP port to listen on.
    """

    taskgroup = None
    single = False

    def __init__(self, identity=None, address=None, port=None):
        super().__init__(identity=identity)

        from pymodbus.framer.socket_framer import (  # pylint: disable=import-outside-toplevel
            ModbusSocketFramer,
        )

        self.decoder = ServerDecoder()
        self.framer = ModbusSocketFramer
        self.address = address or "localhost"
        self.port = (
            port if port is not None else Defaults.Port  # pylint: disable=no-member  # YES IT DOES
        )

    async def serve(self, opened:anyio.Event|None=None):
        """Run this server.
        Sets the `opened` event, if given, as soon as the server port is open.
        """
        try:
            async with anyio.create_task_group() as tg:
                self.taskgroup = tg
                async with await anyio.create_tcp_listener(
                    local_port=self.port,
                    local_host=self.address,
                    reuse_port=True,
                ) as server:
                    if self.port == 0:
                        self.port = server.listeners[0].extra_attributes[
                            SocketAttribute.local_address
                        ]()[1]
                    if opened is not None:
                        opened.set()

                    await server.serve(self._serve_one)
        except socket.gaierror:
            _logger.error("Trying to look up %s", self.address)
            raise
        finally:
            self.taskgroup = None

    async def _serve_one(self, conn):
        reset_frame = False
        framer = self.framer(decoder=self.decoder)

        while True:
            try:
                data = await conn.receive(4096)
                if data == b"":
                    break
                if _logger.isEnabledFor(logging.DEBUG):
                    _logger.debug(  # pylint: disable=logging-not-lazy
                        "Handling data: " + hexlify_packets(data)
                    )

                reqs = []
                # TODO fix pymodbus
                framer.processIncomingPacket(data, unit=0, callback=reqs.append, single=True)

                for request in reqs:
                    unit = request.unit_id
                    tid = request.transaction_id
                    try:
                        response = await self.process_request(request)
                    except NoSuchSlaveException:
                        _logger.debug("requested slave does not exist: %d", request.unit_id)
                        response = request.doException(merror.GatewayNoResponse)
                    except Exception as exc:  # pylint: disable=broad-except
                        _logger.warning("Datastore unable to fulfill request", exc_info=exc)
                        response = request.doException(merror.SlaveFailure)
                    if response.should_respond:
                        response.transaction_id = tid
                        response.unit_id = unit
                        # self.server.control.Counter.BusMessage += 1
                        pdu = framer.buildPacket(response)
                        if _logger.isEnabledFor(logging.DEBUG):
                            _logger.debug("send: %s", b2a_hex(pdu))
                        await conn.send(pdu)

            except socket.timeout as msg:
                _logger.debug("Socket timeout occurred: %r", msg)
                reset_frame = True
            except socket.error as msg:
                _logger.error("Socket error occurred: %r", msg)
                return
            except anyio.get_cancelled_exc_class():
                raise
            except anyio.BrokenResourceError:
                return
            except Exception:  # pylint: disable=broad-except
                _logger.exception("Server error")
                return
            finally:
                if reset_frame:
                    framer.resetFrame()
                    reset_frame = False


class MockAioModbusServer(ModbusServer):
    """A test modbus server with static data"""

    pass
